@@ -20,12 +20,17 @@
 //!
 //! [Unwinding across an FFI boundary is Undefined Behaviour](https://doc.rust-lang.org/nomicon/ffi.html#ffi-and-panics).
 //!
-//! If a the `poll` function of [`FfiFuture`] panics, the unwinding is caught with [`std::panic::catch_unwind`], and then propagated on the host.
+//! ### Panic in `Future::poll`
 //!
-//! This works similarly as [`futures::FutureExt::catch_unwind`](https://docs.rs/futures/0.3.16/futures/future/trait.FutureExt.html#method.catch_unwind),
-//! but instead of returning a `Result` from every future, if the future panics, the host will panic too.
+//! Since the body of `async fn` is translated to `Future::poll` by compiler, it is most likely to
+//! panic. In this case, the wrapped [`FfiFuture`] will catch unwinding with [`std::panic::catch_unwind`],
+//! returning [`FfiPoll::Panicked`]. And the other side (usually the plugin host) will get this value
+//! and explicit panic, just like [`std::sync::Mutex`]'s poisoning mechanism.
 //!
-//! [`std::panic::catch_unwind`]: std::panic::catch_unwind
+//! ### Panic in `Future::drop` or any waker vtable functions `Waker::*`
+//!
+//! Unfortunately, this is very difficult to handle since drop cleanup and `Waker` functions are
+//! expected to be infallible. If these functions panic, we would just call [`std::process::abort`].
 //!
 //! ## Example
 //!
@@ -71,9 +76,8 @@ use std::{
     fmt,
     future::Future,
     marker::PhantomData,
-    mem::ManuallyDrop,
+    mem::{self, ManuallyDrop},
     pin::Pin,
-    process::abort,
     task::{Context, Poll, RawWaker, RawWakerVTable, Waker},
 };
 
@@ -96,6 +100,30 @@ pub enum FfiPoll<T> {
     Panicked,
 }
 
+/// Abort on drop with a message.
+struct DropBomb(&'static str);
+
+impl DropBomb {
+    fn with<T, F: FnOnce() -> T>(message: &'static str, f: F) -> T {
+        let bomb = DropBomb(message);
+        let ret = f();
+        mem::forget(bomb);
+        ret
+    }
+}
+
+impl Drop for DropBomb {
+    fn drop(&mut self) {
+        use std::io::Write;
+        // Use `Stderr::write_all` instead of `eprintln!` to avoid panicking here.
+        let mut stderr = std::io::stderr();
+        let _ = stderr.write_all(b"async-ffi: abort due to panic across the FFI boundary in ");
+        let _ = stderr.write_all(self.0.as_bytes());
+        let _ = stderr.write_all(b"\n");
+        std::process::abort();
+    }
+}
+
 /// The FFI compatible [`std::task::Context`]
 ///
 /// [`std::task::Context`]: std::task::Context
@@ -109,16 +137,21 @@ pub struct FfiContext<'a> {
 }
 
 impl<'a> FfiContext<'a> {
-    fn new(waker: &'a FfiWaker) -> Self {
+    /// SAFETY: VTable functions of `waker` are unsafe, the caller must ensure they have a
+    /// sane behavior as a Waker. `with_context` relies on this to be safe.
+    unsafe fn new(waker: &'a FfiWaker) -> Self {
         Self {
             waker_ref: waker as *const _ as *const FfiWaker,
             phantom: PhantomData,
         }
     }
+
     /// Runs a closure with the [`FfiContext`] as a normal [`std::task::Context`].
     ///
     /// [`std::task::Context`]: std::task::Context
-    pub unsafe fn with_context<T, F: FnOnce(&mut Context) -> T>(&mut self, closure: F) -> T {
+    pub fn with_context<T, F: FnOnce(&mut Context) -> T>(&mut self, closure: F) -> T {
+        // C vtable functions are considered from FFI and they are not expected to unwind, so we don't
+        // need to wrap them with `DropBomb`.
         static RUST_WAKER_VTABLE: RawWakerVTable = {
             unsafe fn clone(data: *const ()) -> RawWaker {
                 let waker = data.cast::<FfiWaker>();
@@ -140,11 +173,15 @@ impl<'a> FfiContext<'a> {
             RawWakerVTable::new(clone, wake, wake_by_ref, drop)
         };
 
-        // The `waker_ref` is borrowed from external context. We must not call drop on it.
-        let waker = ManuallyDrop::new(Waker::from_raw(RawWaker::new(
-            self.waker_ref.cast(),
-            &RUST_WAKER_VTABLE,
-        )));
+        // SAFETY: `waker_ref`'s vtable functions must have behavior, this is the contract of
+        // `FfiContext::new`.
+        let waker = unsafe {
+            // The `waker_ref` is borrowed from external context. We must not call drop on it.
+            ManuallyDrop::new(Waker::from_raw(RawWaker::new(
+                self.waker_ref.cast(),
+                &RUST_WAKER_VTABLE,
+            )))
+        };
         let mut ctx = Context::from_waker(&*waker);
 
         closure(&mut ctx)
@@ -165,29 +202,37 @@ impl<'a> ContextExt for Context<'a> {
     fn with_ffi_context<T, F: FnOnce(&mut FfiContext) -> T>(&mut self, closure: F) -> T {
         static C_WAKER_VTABLE_OWNED: FfiWakerVTable = {
             unsafe extern "C" fn clone(data: *const FfiWaker) -> *const FfiWaker {
-                let waker: Waker = (*(*data).waker.owned).clone();
-                Box::into_raw(Box::new(FfiWaker {
-                    vtable: &C_WAKER_VTABLE_OWNED,
-                    waker: WakerUnion {
-                        owned: ManuallyDrop::new(waker),
-                    },
-                }))
-                .cast()
+                DropBomb::with("Waker::clone", || {
+                    let waker: Waker = (*(*data).waker.owned).clone();
+                    Box::into_raw(Box::new(FfiWaker {
+                        vtable: &C_WAKER_VTABLE_OWNED,
+                        waker: WakerUnion {
+                            owned: ManuallyDrop::new(waker),
+                        },
+                    }))
+                    .cast()
+                })
             }
             // In this case, we must own `data`. This can only happen when the `data` pointer is returned from `clone`.
             // Thus the it is `Box<FfiWaker>`.
             unsafe extern "C" fn wake(data: *const FfiWaker) {
-                let b = Box::from_raw(data as *mut FfiWaker);
-                ManuallyDrop::into_inner(b.waker.owned).wake();
+                DropBomb::with("Waker::wake", || {
+                    let b = Box::from_raw(data as *mut FfiWaker);
+                    ManuallyDrop::into_inner(b.waker.owned).wake();
+                })
             }
             unsafe extern "C" fn wake_by_ref(data: *const FfiWaker) {
-                (*data).waker.owned.wake_by_ref();
+                DropBomb::with("Waker::wake_by_ref", || {
+                    (*data).waker.owned.wake_by_ref();
+                })
             }
             // Same as `wake`.
             unsafe extern "C" fn drop(data: *const FfiWaker) {
-                let mut b = Box::from_raw(data as *mut FfiWaker);
-                ManuallyDrop::drop(&mut b.waker.owned);
-                std::mem::drop(b);
+                DropBomb::with("Waker::drop", || {
+                    let mut b = Box::from_raw(data as *mut FfiWaker);
+                    ManuallyDrop::drop(&mut b.waker.owned);
+                    mem::drop(b);
+                });
             }
             FfiWakerVTable {
                 clone,
@@ -199,20 +244,24 @@ impl<'a> ContextExt for Context<'a> {
 
         static C_WAKER_VTABLE_REF: FfiWakerVTable = {
             unsafe extern "C" fn clone(data: *const FfiWaker) -> *const FfiWaker {
-                let waker: Waker = (*(*data).waker.reference).clone();
-                Box::into_raw(Box::new(FfiWaker {
-                    vtable: &C_WAKER_VTABLE_OWNED,
-                    waker: WakerUnion {
-                        owned: ManuallyDrop::new(waker),
-                    },
-                }))
-                .cast()
+                DropBomb::with("Waker::clone", || {
+                    let waker: Waker = (*(*data).waker.reference).clone();
+                    Box::into_raw(Box::new(FfiWaker {
+                        vtable: &C_WAKER_VTABLE_OWNED,
+                        waker: WakerUnion {
+                            owned: ManuallyDrop::new(waker),
+                        },
+                    }))
+                    .cast()
+                })
             }
             unsafe extern "C" fn wake_by_ref(data: *const FfiWaker) {
-                (*(*data).waker.reference).wake_by_ref();
+                DropBomb::with("Waker::wake_by_ref", || {
+                    (*(*data).waker.reference).wake_by_ref();
+                })
             }
             unsafe extern "C" fn unreachable(_: *const FfiWaker) {
-                abort();
+                std::process::abort();
             }
             FfiWakerVTable {
                 clone,
@@ -229,7 +278,9 @@ impl<'a> ContextExt for Context<'a> {
             },
         };
 
-        let mut ctx = FfiContext::new(&waker);
+        // SAFETY: The behavior of `waker` is sane since we forward them to another valid Waker.
+        // That waker must be safe to use due to the contract of `RawWaker::new`.
+        let mut ctx = unsafe { FfiContext::new(&waker) };
 
         closure(&mut ctx)
     }
@@ -357,7 +408,7 @@ impl<'a, T> BorrowingFfiFuture<'a, T> {
     }
 }
 
-// This is safe since we allow only `Send` Future in `FfiFuture::new`.
+// SAFETY: This is safe since we allow only `Send` Future in `FfiFuture::new`.
 unsafe impl<T> Send for BorrowingFfiFuture<'_, T> {}
 
 impl<T> Future for BorrowingFfiFuture<'_, T> {
@@ -399,23 +450,27 @@ impl<'a, T> LocalBorrowingFfiFuture<'a, T> {
             fut_ptr: *mut (),
             context_ptr: *mut FfiContext,
         ) -> FfiPoll<F::Output> {
-            let fut_pin = Pin::new_unchecked(&mut *fut_ptr.cast::<F>());
-            (*context_ptr).with_context(|ctx| {
-                // Unwinding across an FFI boundary is UB
-                // https://doc.rust-lang.org/nomicon/ffi.html#ffi-and-panics
-                //
-                // so we catch the unwind, if any occurs, and then panic on the host.
-                match std::panic::catch_unwind(std::panic::AssertUnwindSafe(|| {
-                    F::poll(fut_pin, ctx)
-                })) {
-                    Ok(p) => p.into(),
-                    Err(_) => FfiPoll::Panicked,
+            // The poll fn is likely to panic since it contains most of user logic.
+            match std::panic::catch_unwind(std::panic::AssertUnwindSafe(|| {
+                let fut_pin = Pin::new_unchecked(&mut *fut_ptr.cast::<F>());
+                (*context_ptr).with_context(|ctx| F::poll(fut_pin, ctx))
+            })) {
+                Ok(p) => p.into(),
+                Err(payload) => {
+                    // Panic payload may panic when dropped, ensure not propagate it.
+                    // https://github.com/rust-lang/rust/issues/86027
+                    DropBomb::with("drop of panic payload from Future::poll", move || {
+                        drop(payload);
+                    });
+                    FfiPoll::Panicked
                 }
-            })
+            }
         }
 
         unsafe extern "C" fn drop_fn<T>(ptr: *mut ()) {
-            drop(Box::from_raw(ptr.cast::<T>()));
+            DropBomb::with("Future::drop", || {
+                drop(Box::from_raw(ptr.cast::<T>()));
+            });
         }
 
         let ptr = Box::into_raw(Box::new(fut));
@@ -430,6 +485,10 @@ impl<'a, T> LocalBorrowingFfiFuture<'a, T> {
 
 impl<T> Drop for LocalBorrowingFfiFuture<'_, T> {
     fn drop(&mut self) {
+        // SAFETY: This is safe since `drop_fn` is construct from `LocalBorrowingFfiFuture::new`
+        // and is a dropper
+        // `LocalBorrowingFfiFuture::new` and they are just a Box pointer and its coresponding
+        // dropper.
         unsafe { (self.drop_fn)(self.fut_ptr) };
     }
 }
@@ -438,6 +497,8 @@ impl<T> Future for LocalBorrowingFfiFuture<'_, T> {
     type Output = T;
 
     fn poll(self: Pin<&mut Self>, ctx: &mut Context<'_>) -> Poll<Self::Output> {
+        // SAFETY: This is safe since `poll_fn` is constructed from `LocalBorrowingFfiFuture::new`
+        // and it just forwards to the original safe `Future::poll`.
         ctx.with_ffi_context(|ctx| unsafe { (self.poll_fn)(self.fut_ptr, ctx) })
             .try_into()
             // Propagate panic from FFI.
